@@ -1,15 +1,19 @@
 
-import os, pickle, time, bz2
+import os, sys, pickle, time, bz2
 from cStringIO import StringIO
 from zope.interface import implements
 from twisted.trial import unittest
 from twisted.application import service
 from twisted.internet import defer
+try:
+    from twisted import logger as twisted_logger
+except ImportError:
+    twisted_logger = None
 from twisted.python import log as twisted_log
 from twisted.python import failure, runtime, usage
 import foolscap
 from foolscap.logging import gatherer, log, tail, incident, cli, web, \
-     publish, dumper
+     publish, dumper, flogfile
 from foolscap.logging.interfaces import RILogObserver
 from foolscap.util import format_time
 from foolscap.eventual import fireEventually, flushEventualQueue
@@ -195,6 +199,97 @@ class Advanced(unittest.TestCase):
         n2 = l.msg("two", parent=n)
         l.msg("three", parent=n2)
 
+class ErrorfulQualifier(incident.IncidentQualifier):
+    def __init__(self):
+        self._first = True
+
+    def check_event(self, ev):
+        if self._first:
+            self._first = False
+            raise ValueError("oops")
+        return False
+
+class NoStdio(unittest.TestCase):
+    # bug #244 is caused, in part, by Foolscap-side logging failures which
+    # write an error message ("unable to serialize X") to stderr, which then
+    # gets captured by twisted's logging (when run in a program under
+    # twistd), then fed back into foolscap logging. Check that unserializable
+    # objects don't cause anything to be written to a mock stdout/stderr
+    # object.
+    #
+    # FoolscapLogger used stdio in two places:
+    # * msg() when format_message() throws
+    # * add_event() when IncidentQualifier.event() throws
+
+    def setUp(self):
+        self.fl = log.FoolscapLogger()
+        self.mock_stdout = StringIO()
+        self.mock_stderr = StringIO()
+        self.orig_stdout = sys.stdout
+        self.orig_stderr = sys.stderr
+        sys.stdout = self.mock_stdout
+        sys.stderr = self.mock_stderr
+
+    def tearDown(self):
+        sys.stdout = self.orig_stdout
+        sys.stderr = self.orig_stderr
+
+    def check_stdio(self):
+        self.failUnlessEqual(self.mock_stdout.getvalue(), "")
+        self.failUnlessEqual(self.mock_stderr.getvalue(), "")
+
+    def test_unformattable(self):
+        self.fl.msg(format="one=%(unformattable)s") # missing format key
+        self.check_stdio()
+
+    def test_unserializable_incident(self):
+        # one #244 pathway involved an unserializable event that caused an
+        # exception during IncidentReporter.incident_declared(), as it tried
+        # to record all recent events. We can test the lack of stdio by using
+        # a qualifier that throws an error directly.
+        self.fl.setIncidentQualifier(ErrorfulQualifier())
+        self.fl.activate_incident_qualifier()
+        # make sure we set it up correctly
+        self.failUnless(self.fl.active_incident_qualifier)
+        self.fl.msg("oops", arg=lambda : "lambdas are unserializable",
+                    level=log.BAD)
+        self.check_stdio()
+        # The internal error will cause a new "metaevent" to be recorded. The
+        # original event may or may not get recorded first, depending upon
+        # the error (i.e. does it happen before or after buffer.append is
+        # called). Also, get_buffered_events() is unordered. So search for
+        # the right one.
+        events = [e for e in self.fl.get_buffered_events()
+                  if e.get("facility") == "foolscap/internal-error"]
+        self.assertEqual(len(events), 1)
+        m = events[0]["message"]
+        expected = "internal error in log._msg, args=('oops',)"
+        self.assert_(m.startswith(expected), m)
+        self.assertIn("ValueError('oops'", m)
+
+
+class Serialization(unittest.TestCase):
+    def test_lazy_serialization(self):
+        # Both foolscap and twisted allow (somewhat) arbitrary kwargs in the
+        # log.msg() call. Twisted will either discard the event (if nobody is
+        # listening), or stringify it right away.
+        #
+        # Foolscap does neither. It records the event (kwargs and all) in a
+        # circular buffer, so a later observer can learn about them (either
+        # 'flogtool tail' or a stored Incident file). And it stores the
+        # arguments verbatim, leaving stringification to the future observer
+        # (if they want it), so tools can filter events without using regexps
+        # or parsing prematurely-flattened strings.
+        #
+        # Test this by logging a mutable object, modifying it, then checking
+        # the buffer. We expect to see the modification.
+        fl = log.FoolscapLogger()
+        mutable = {"key": "old"}
+        fl.msg("one", arg=mutable)
+        mutable["key"] = "new"
+        events = list(fl.get_buffered_events())
+        self.failUnless(events[0]["arg"]["key"], "new")
+
 class SuperstitiousQualifier(incident.IncidentQualifier):
     def check_event(self, ev):
         if "thirteen" in ev.get("message", ""):
@@ -210,20 +305,7 @@ class NoFollowUpReporter(incident.IncidentReporter):
 
 class LogfileReaderMixin:
     def _read_logfile(self, fn):
-        if fn.endswith(".bz2"):
-            f = bz2.BZ2File(fn, "r")
-        else:
-            f = open(fn, "rb")
-        events = []
-        while True:
-            try:
-                events.append(pickle.load(f))
-            except EOFError:
-                break
-            except ValueError:
-                break
-        f.close()
-        return events
+        return list(flogfile.get_events(fn, ignore_value_error=True))
 
 class Incidents(unittest.TestCase, PollMixin, LogfileReaderMixin):
     def test_basic(self):
@@ -539,14 +621,15 @@ class Publish(PollMixin, unittest.TestCase):
             def _check_observer(res):
                 msgs = ob.messages
                 self.failUnlessEqual(len(msgs), 8)
-                #print msgs
                 self.failUnlessEqual(msgs[0]["message"], "message 1 here")
+                self.failUnlessEqual(msgs[1]["from-twisted"], True)
                 self.failUnlessEqual(msgs[1]["message"], "message 2 here")
                 self.failUnlessEqual(msgs[1]["tubID"], t.tubID)
+                self.failUnlessEqual(msgs[2]["from-twisted"], True)
                 self.failUnlessEqual(msgs[2]["message"], "message 3 here")
                 self.failUnlessEqual(msgs[2]["tubID"], None)
-                self.failUnlessEqual(msgs[3]["format"], "%(foo)s is foo")
-                self.failUnlessEqual(msgs[3]["foo"], "foo")
+                self.failUnlessEqual(msgs[3]["from-twisted"], True)
+                self.failUnlessEqual(msgs[3]["message"], "foo is foo")
 
                 # check the errors
                 self.failUnlessEqual(msgs[4]["message"], "")
@@ -560,25 +643,16 @@ class Publish(PollMixin, unittest.TestCase):
                 self.failUnless(msgs[5]["failure"].check(SampleError))
                 self.failUnless("err2" in str(msgs[5]["failure"]))
 
-                # twisted-8.0 has textFromEventDict, which means we get a
-                # ["message"] key from log.err . In older version of
-                # twisted, we don't.
-                if msgs[6]["message"]:
-                    self.failUnless("Unhandled Error" in msgs[6]["message"])
-                    self.failUnless("SampleError: err3" in msgs[6]["message"])
+                # errors coming from twisted are stringified
+                self.failUnlessEqual(msgs[6]["from-twisted"], True)
+                self.failUnless("Unhandled Error" in msgs[6]["message"])
+                self.failUnless("SampleError: err3" in msgs[6]["message"])
                 self.failUnless(msgs[6]["isError"])
-                self.failUnless("failure" in msgs[6])
-                self.failUnless(msgs[6]["failure"].check(SampleError))
-                self.failUnless("err3" in str(msgs[6]["failure"]))
 
-                # same
-                if msgs[7]["message"]:
-                    self.failUnless("Unhandled Error" in msgs[7]["message"])
-                    self.failUnless("SampleError: err4" in msgs[7]["message"])
+                self.failUnlessEqual(msgs[7]["from-twisted"], True)
+                self.failUnless("Unhandled Error" in msgs[7]["message"])
+                self.failUnless("SampleError: err4" in msgs[7]["message"])
                 self.failUnless(msgs[7]["isError"])
-                self.failUnless("failure" in msgs[7])
-                self.failUnless(msgs[7]["failure"].check(SampleError))
-                self.failUnless("err4" in str(msgs[7]["failure"]))
 
             d.addCallback(_check_observer)
             def _done(res):
@@ -1546,7 +1620,7 @@ class Tail(unittest.TestCase):
                        })
         outmsg = out.getvalue()
         # this contains a localtime string, so don't check the hour
-        self.failUnless(":25:06.527 L25 []#123 howdy" in outmsg)
+        self.failUnless(":06.527 L25 []#123 howdy" in outmsg)
 
         lp.remote_msg({"time": 1207005907.527782,
                        "level": 25,
@@ -1556,7 +1630,7 @@ class Tail(unittest.TestCase):
                        })
         outmsg = out.getvalue()
         # this contains a localtime string, so don't check the hour
-        self.failUnless(":25:07.527 L25 []#124 howdy pardner" in outmsg)
+        self.failUnless(":07.527 L25 []#124 howdy pardner" in outmsg)
 
         try:
             raise RuntimeError("fake error")
@@ -1571,7 +1645,7 @@ class Tail(unittest.TestCase):
                        })
         outmsg = out.getvalue()
 
-        self.failUnless(":25:50.002 L30 []#125 oops\n FAILURE:\n" in outmsg,
+        self.failUnless(":50.002 L30 []#125 oops\n FAILURE:\n" in outmsg,
                         outmsg)
         self.failUnless("exceptions.RuntimeError" in outmsg, outmsg)
         self.failUnless(": fake error" in outmsg, outmsg)
@@ -2130,22 +2204,79 @@ class Bridge(unittest.TestCase):
 
         tw.msg("one")
         tw.msg(format="two %(two)d", two=2)
+        # twisted now has places (e.g. Factory.doStart) where the new
+        # Logger.info() is called with arbitrary (unserializable) kwargs for
+        # string formatting, which are passed into the old LogPublisher(),
+        # from which they arrive in foolscap. Make sure we can tolerate that.
+        # The rule is that foolscap immediately stringifies all events it
+        # gets from twisted (with log.textFromEventDict), and doesn't store
+        # the additional arguments. So it's ok to put an *unserializable*
+        # argument into the log.msg() call, as long as it's still
+        # *stringifyable*.
+        unserializable = lambda: "unserializable"
+        tw.msg(format="three is %(evil)s", evil=unserializable)
+
         d = flushEventualQueue()
         def _check(res):
-            self.failUnlessEqual(len(tw_out), 2)
+            self.failUnlessEqual(len(tw_out), 3)
             self.failUnlessEqual(tw_out[0]["message"], ("one",))
             self.failUnlessEqual(tw_out[1]["format"], "two %(two)d")
             self.failUnlessEqual(tw_out[1]["two"], 2)
-
-            self.failUnlessEqual(len(fl_out), 2)
+            self.failUnlessEqual(tw_out[2]["format"], "three is %(evil)s")
+            self.failUnlessEqual(tw_out[2]["evil"], unserializable)
+            self.failUnlessEqual(len(fl_out), 3)
             self.failUnlessEqual(fl_out[0]["message"], "one")
             self.failUnless(fl_out[0]["from-twisted"])
-            self.failUnlessEqual(fl_out[1]["format"], "two %(two)d")
+            self.failUnlessEqual(fl_out[1]["message"], "two 2")
             self.failUnless(fl_out[1]["from-twisted"])
+            # str(unserializable) is like "<function <lambda> at 0xblahblah>"
+            self.failUnlessEqual(fl_out[2]["message"],
+                                 "three is " + str(unserializable))
+            self.failUnless(fl_out[2]["from-twisted"])
 
         d.addCallback(_check)
         return d
 
+    def test_twisted_logger_to_foolscap(self):
+        if not twisted_logger:
+            raise unittest.SkipTest("needs twisted.logger from Twisted>=15.2.0")
+        new_pub = twisted_logger.LogPublisher()
+        old_pub = twisted_log.LogPublisher(observerPublisher=new_pub,
+                                           publishPublisher=new_pub)
+        fl = log.FoolscapLogger()
+        log.bridgeLogsFromTwisted(None, old_pub, fl)
+        tw_out = []
+        old_pub.addObserver(tw_out.append)
+        fl_out = []
+        fl.addObserver(fl_out.append)
+
+        tl = twisted_logger.Logger(observer=new_pub)
+        tl.info("one")
+        # note: new twisted logger wants PEP3101 format strings, {} not %
+        tl.info(format="two {two}", two=2)
+        # twisted's new Logger.info() takes arbitrary (unserializable) kwargs
+        # for string formatting, and passes them into the old LogPublisher(),
+        # so make sure we can tolerate that. The rule is that foolscap
+        # stringifies all events it gets from twisted, and doesn't store the
+        # additional arguments.
+        unserializable = lambda: "unserializable"
+        tl.info("three is {evil!s}", evil=unserializable)
+
+        d = flushEventualQueue()
+        def _check(res):
+            self.failUnlessEqual(len(fl_out), 3)
+            self.failUnlessEqual(fl_out[0]["message"], "one")
+            self.failUnless(fl_out[0]["from-twisted"])
+            self.failUnlessEqual(fl_out[1]["message"], "two 2")
+            self.failIf("two" in fl_out[1])
+            self.failUnless(fl_out[1]["from-twisted"])
+            # str(unserializable) is like "<function <lambda> at 0xblahblah>"
+            self.failUnlessEqual(fl_out[2]["message"],
+                                 "three is " + str(unserializable))
+            self.failUnless(fl_out[2]["from-twisted"])
+
+        d.addCallback(_check)
+        return d
 
     def test_no_loops(self):
         fl = log.FoolscapLogger()
